/*
 *  RecAgent.h
 *  
 *  Created by Jordan Frank on Thu December 7 2006.
 *
 *  Description: 
 *  
 *	Designed for the PudleWorld.cpp environment
 *
 *  This agent uses off-policy temporal-difference learning with function approximation
 *  and recognizers, from [1].
 * 
 *  Uses CMAC tile coding based on TileAgent.cpp from RL-Glue 
 *
 *	State: Continous valued - 3D
 *	Actions: Continous Valued - 2D
 *
 *  [1] - Precup, D., Sutton, R. S., Paduraru, C., Koop, A., Singh, S., 
 *        Off-policy Learning with Recognizers, 2006.
 */
 
#ifndef RECAGENT_H
#define RECAGENT_H

#include <iostream>
#include <cstdlib>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "RLcommon.h"
#include "C_agent_parser.h"
#include "tiles.h"

// Global vars********************************************************************************
int stateDim;       // number of state variables
int numActions;     // number of actions
int step;

double *theta;              // learnable weight vector for reward estimates
double *non_rec_theta;      // We want to compare our estimates with
                            // an estimate that uses an exact rec. prob.
double *mu_weights;         // learned weight vector for recognition probabilities

double k;
double non_rec_k;           // for exact rec. prob. case.

double *e;                  // eligibility trace vector
double *non_rec_e;          // for exact rec. prob. case.

const double lambda = 0.6;
const double beta = 0.005;         // termination condition, just a constant for now
const double g = 0.005;            // restart function, just a constant for now

double alpha = 0.001;      // learning rate
double mu_alpha = 0.005;   // learning rate for rec. probs.

extern const int T = 100000;       // Number of time steps per round. Used to decay learning rates.
int t;                     // Current time step.

const int memorysize = 512;      // mem size for tile coding
const int tilings = 8;           // number of tiling sheets
const double tilewidth = 0.125;    // scale factor for tiles
float *input;               // [state,action] given to tile coder
int *featurevec;            // feature vector for a state

Observation prev_state;     // old state value
Action prev_action;         // old action value
Observation curr_state;     // current state value
Action curr_action;         // current action

// Global methods*****************************************************************************
Action randomAction();
Action favourUpRightAction();
Action greedyAction(Observation);
int argmax(double *);
void getFeatureVector(Observation, int* phi);

// Declare a Recognizer as a function that takes an Observation and an Action and returns
// a 1 or a 0 if the Action is recognized in the Observation or not.
typedef int (*Recognizer)(Observation, Action);
int recognize(Recognizer, Observation, Action);
int recognize_up_right(Observation, Action);
int recognize_good_actions(Observation, Action);
double getExactComplexRecProb(Observation);

// For logistic regression learning for recognition probabilities
double getMu(int *);
void updateMu(int *, double);
double vecvecMult(int *, double *, int);
double sigmoid(double);
void decayLearningRates(int t);
int recognized;
int total;

#endif
